#include "Memory.h"

#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/libraries/armem/server/wm/memory_definitions.h>

// ArmarX
#include "../base/filter/frequencyFilter/FrequencyFilter.h"
#include "../base/extractor/noExtractor/NoExtractor.h"
#include "../base/extractor/imageExtractor/ImageExtractor.h"
#include "../base/converter/dict/json/JsonConverter.h"
#include "../base/converter/dict/bson/BsonConverter.h"
#include "../base/converter/image/png/PngConverter.h"

namespace armarx::armem::server::ltm::disk
{
    namespace
    {
        MemoryID getMemoryIDFromPath(const std::filesystem::path& p)
        {
            ARMARX_CHECK(!p.empty());

            MemoryID m;
            m.memoryName = p.filename();
            return m;
        }

        std::filesystem::path getDefaultStoragePath()
        {
            /*std::string armarx_home = std::string(getenv("HOME")) + "/.armarx";
            if (getenv("ARMARX_DEFAULTS_DIR"))
            {
                armarx_home = getenv("ARMARX_DEFAULTS_DIR");
            }
            path = armarx_home + "/armem/disk/data/db";*/
            return "/tmp/MemoryExport/Test";
        }
    }

    void Memory::createPropertyDefinitions(PropertyDefinitionsPtr& properties, const std::string& prefix)
    {
        properties->optional(path, prefix + "storagepath", "The path to the memory storage.");
    }

    Memory::Memory() :
        Memory(getDefaultStoragePath())
    {
    }

    Memory::Memory(const std::filesystem::path& p) :
        Base(getMemoryIDFromPath(p)),
        BufferedMemoryItem(getMemoryIDFromPath(p)),
        DiskMemoryItem(p)
    {
        pipeline.memFilter = std::make_shared<filter::MemoryFrequencyFilter>();
        pipeline.snapFilter = std::make_shared<filter::SnapshotFrequencyFilter>();
        pipeline.extractor = std::make_shared<extractor::ImageExtractor>();
        pipeline.dictConverter = std::make_shared<converter::dict::JsonConverter>();
        pipeline.imgConverter = std::make_shared<converter::image::PngConverter>();
    }

    void Memory::setMemoryID(const MemoryID& id)
    {
        ARMARX_CHECK_NOT_EMPTY(id.memoryName);

        Base::setMemoryID(id.getMemoryID());
        std::filesystem::path p(path);
        path = (p.parent_path() / id.memoryName);
        buffer.id() = id.getMemoryID();
    }

    std::string Memory::getExpectedFolderName() const
    {
        return id().memoryName;
    }

    bool Memory::forEachCoreSegment(std::function<void(CoreSegment&)>&& func) const
    {
        if (!checkPath())
        {
            return false;
        }

        std::filesystem::path p = path;
        util::ensureFolderExists(p, false);

        for (const auto& subdir : std::filesystem::directory_iterator(p))
        {
            std::filesystem::path subdirPath = subdir.path();
            CoreSegment c(subdirPath, pipeline);
            func(c);
        }
        return true;
    }

    std::shared_ptr<CoreSegment> Memory::findCoreSegment(const std::string& n) const
    {
        if (!checkPath())
        {
            return {};
        }

        std::filesystem::path p = path;
        util::ensureFolderExists(p, false);

        std::filesystem::path subpath = p / n;
        util::ensureFolderExists(subpath, false);
        auto c = std::make_shared<CoreSegment>(subpath, pipeline);
        return c;
    }

    void Memory::_loadAll(armem::wm::Memory& m)
    {
        m.id() = id();

        forEachCoreSegment([&m](CoreSegment& x) {
            armem::wm::CoreSegment s;
            x.loadAll(s);
            m.addCoreSegment(s);
        });
    }

    void Memory::_load(armem::wm::Memory& m)
    {
        if (!checkPath())
        {
            return;
        }

        std::lock_guard l(ltm_mutex);
        m.forEachCoreSegment([this](armem::wm::CoreSegment& e)
        {
            util::ensureFolderExists(std::filesystem::path(path) / e.id().coreSegmentName, false);
            CoreSegment c((std::filesystem::path(path) / e.id().coreSegmentName), pipeline);
            c.load(e);
        });
    }

    void Memory::_store(const armem::wm::Memory& memory)
    {
        std::lock_guard l(this->ltm_mutex);
        buffer.append(memory);
    }

    void Memory::_directlyStore(const armem::wm::Memory& memory)
    {
        if (!this->pipeline.memFilter->accept(memory))
        {
            ARMARX_WARNING << deactivateSpam() << "Ignoring to put a Memory into the LTM because it got filtered.";
            return;
        }

        if (!checkPath())
        {
            return;
        }

        std::lock_guard l(ltm_mutex);

        memory.forEachCoreSegment([this](const auto& core)
        {
            util::ensureFolderExists(std::filesystem::path(path) / core.id().coreSegmentName);
            CoreSegment c((std::filesystem::path(path) / core.id().coreSegmentName), pipeline);
            c.store(core);
        });
    }

}
