/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/type/variant/Variant.h>
#include <RobotAPI/libraries/aron/core/codegenerator/helper/ReaderInfo.h>
#include <RobotAPI/libraries/aron/core/codegenerator/helper/WriterInfo.h>

#include <ArmarXCore/libraries/cppgen/CppBlock.h>
#include <ArmarXCore/libraries/cppgen/CppField.h>
#include <ArmarXCore/libraries/cppgen/CppCtor.h>
#include <ArmarXCore/libraries/cppgen/CppMethod.h>
#include <ArmarXCore/libraries/cppgen/CppClass.h>

#include <memory>
#include <map>
#include <string>
#include <vector>


namespace armarx::aron::codegenerator::cpp
{

    namespace conversion
    {
        const std::map<type::Maybe, std::string> Maybe2CppString =
        {
            {type::Maybe::eNone, "armarx::aron::type::Maybe::eNone"},
            {type::Maybe::eOptional, "armarx::aron::type::Maybe::eOptional"},
            {type::Maybe::eRawPointer, "armarx::aron::type::Maybe::eRawPointer"},
            {type::Maybe::eSharedPointer, "armarx::aron::type::Maybe::eSharedPointer"},
            {type::Maybe::eUniquePointer, "armarx::aron::type::Maybe::eUniquePointer"}
        };

        const std::map<type::matrix::ElementType, std::string> MatrixElementType2CppString =
        {
            {type::matrix::int16, "::armarx::aron::type::matrix::int16"},
            {type::matrix::int32, "::armarx::aron::type::matrix::int32"},
            {type::matrix::int64, "::armarx::aron::type::matrix::int64"},
            {type::matrix::float32, "::armarx::aron::type::matrix::float32"},
            {type::matrix::float64, "::armarx::aron::type::matrix::float64"}
        };

        const std::map<type::pointcloud::VoxelType, std::string> VoxelType2CppString =
        {
            {type::pointcloud::VoxelType::PointXYZ, "::armarx::aron::type::pointcloud::VoxelType::PointXYZ"},
            {type::pointcloud::VoxelType::PointXYZI, "::armarx::aron::type::pointcloud::VoxelType::PointXYZI"},
            {type::pointcloud::VoxelType::PointXYZL, "::armarx::aron::type::pointcloud::VoxelType::PointXYZL"},
            {type::pointcloud::VoxelType::PointXYZRGB, "::armarx::aron::type::pointcloud::VoxelType::PointXYZRGB"},
            {type::pointcloud::VoxelType::PointXYZRGBL, "::armarx::aron::type::pointcloud::VoxelType::PointXYZRGBL"},
            {type::pointcloud::VoxelType::PointXYZRGBA, "::armarx::aron::type::pointcloud::VoxelType::PointXYZRGBA"},
            {type::pointcloud::VoxelType::PointXYZHSV, "::armarx::aron::type::pointcloud::VoxelType::PointXYZHSV"}
        };

        const std::map<type::image::PixelType, std::string> PixelType2CppString =
        {
            {type::image::PixelType::rgb24, "::armarx::aron::type::image::PixelType::rgb24"},
            {type::image::PixelType::depth32, "::armarx::aron::type::image::PixelType::depth32"}
        };

        const std::map<type::quaternion::ElementType, std::string> QuaternionElementType2Cpp =
        {
            {type::quaternion::ElementType::float32, "::armarx::aron::type::quaternion::ElementType::float32"},
            {type::quaternion::ElementType::float64, "::armarx::aron::type::quaternion::ElementType::float64"}
        };
    }

    class GeneratorFactory;
    typedef std::shared_ptr<GeneratorFactory> SerializerFactoryPtr;

    class Generator;
    typedef std::shared_ptr<Generator> GeneratorPtr;

    class Generator
    {
    public:
        using PointerType = GeneratorPtr;

    public:
        // constructors
        Generator() = delete;
        Generator(const std::string& cppName, const std::string& aronDataTypename, const std::string& aronTypeTypename);
        virtual ~Generator() = default;

        // public member methods
        std::string getCoreCppTypename() const;
        std::string getFullCppTypename() const;

        CppMethodPtr toSpecializedDataWriterMethod(const WriterInfo& info) const;
        CppMethodPtr toSpecializedDataReaderMethod(const ReaderInfo& info) const;
        CppMethodPtr toSpecializedStaticDataReaderMethod(const StaticReaderInfo& info) const;
        CppMethodPtr toSpecializedTypeWriterMethod(const WriterInfo& info) const;

        // virtual override definitions
        virtual std::vector<CppFieldPtr> getPublicVariableDeclarations(const std::string&) const;

        CppCtorPtr toCtor(const std::string&) const;
        virtual std::vector<std::pair<std::string, std::string>> getCtorInitializers(const std::string&) const;
        virtual CppBlockPtr getCtorBlock(const std::string&) const;

        CppMethodPtr toResetSoftMethod() const;
        virtual CppBlockPtr getResetSoftBlock(const std::string& cppAccessor) const;

        CppMethodPtr toResetHardMethod() const;
        virtual CppBlockPtr getResetHardBlock(const std::string& cppAccessor) const;

        CppMethodPtr toWriteTypeMethod() const;
        virtual CppBlockPtr getWriteTypeBlock(const std::string& typeAccessor, const std::string& cppAccessor, std::string& variantAccessor) const = 0;

        CppMethodPtr toWriteMethod() const;
        virtual CppBlockPtr getWriteBlock(const std::string& cppAccessor, std::string& variantAccessor) const = 0;

        CppMethodPtr toReadMethod() const;
        virtual CppBlockPtr getReadBlock(const std::string& cppAccessor, const std::string& variantAccessor) const = 0;

        CppMethodPtr toEqualsMethod() const;
        virtual CppBlockPtr getEqualsBlock(const std::string& cppAccessorThis, const std::string& cppAccessorOther) const;

        virtual const type::Variant& getType() const = 0;

        // static methods
        static std::string EscapeAccessor(const std::string&);
        static std::string UnescapeAccessor(const std::string&);

        static std::string ExtractCppTypename(const type::Variant&);
        static std::vector<std::string> ExtractCppTypenames(const std::vector<type::VariantPtr>&);

        static std::unique_ptr<Generator> FromAronType(const type::Variant&);

    protected:
        std::string nextEl() const;
        std::string toPointerAccessor(const std::string&) const;

        std::string resolveMaybeAccessor(const std::string&) const;
        std::string resolveMaybeGenerator(const std::string&) const;

        CppBlockPtr resolveMaybeResetHardBlock(const CppBlockPtr&, const std::string&) const;
        CppBlockPtr resolveMaybeResetSoftBlock(const CppBlockPtr&, const std::string&) const;
        CppBlockPtr resolveMaybeWriteBlock(const CppBlockPtr&, const std::string&) const;
        CppBlockPtr resolveMaybeReadBlock(const CppBlockPtr&, const std::string&, const std::string&) const;
        CppBlockPtr resolveMaybeEqualsBlock(const CppBlockPtr&, const std::string&, const std::string&) const;

    protected:
        static const std::string ARON_VARIABLE_PREFIX;

        static const std::string ARON_MAYBE_TYPE_ACCESSOR;
        static const std::string ARON_READER_ACCESSOR;
        static const std::string ARON_WRITER_ACCESSOR;
        static const std::string ARON_VARIANT_RETURN_ACCESSOR;

    private:
        static const SerializerFactoryPtr FACTORY;

        std::string cppTypename;
        std::string aronDataTypename;
        std::string aronTypeTypename;
    };
}
