/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "ObjectClass.h"

#include <SimoxUtility/meta/type_name.h>


namespace armarx::aron::codegenerator::cpp::generator
{
    // constructors
    ObjectClass::ObjectClass(const type::Object& e) :
        detail::SpecializedGeneratorBase<type::Object, ObjectClass>(e.getObjectName(), simox::meta::get_type_name<data::dto::Dict>(), simox::meta::get_type_name<type::dto::AronObject>(), e)
    {
        if (type.getMaybe() != type::Maybe::eNone)
        {
            throw error::ValueNotValidException(__PRETTY_FUNCTION__, "Somehow the maybe flag of a top level object declaration is set. This is not valid!", std::to_string((int) type.getMaybe()) + " aka " + type::defaultconversion::string::Maybe2String.at(type.getMaybe()), type.getPath());
        }
    }

    std::vector<CppFieldPtr> ObjectClass::getPublicVariableDeclarations(const std::string&) const
    {
        std::vector<CppFieldPtr> fields;
        for (const auto& [key, member] : type.getMemberTypes())
        {
            auto member_s = FromAronType(*member);
            std::vector<CppFieldPtr> member_fields = member_s->getPublicVariableDeclarations(key);
            fields.insert(fields.end(), member_fields.begin(), member_fields.end());
        }
        return fields;
    }

    CppBlockPtr ObjectClass::getResetSoftBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = std::make_shared<CppBlock>();
        if (type.getExtends() != nullptr)
        {
            const auto extends_s = FromAronType(*type.getExtends());
            block_if_data->addLine(extends_s->getFullCppTypename() + "::resetSoft();");
        }

        for (const auto& [key, child] : type.getMemberTypes())
        {
            auto child_s = FromAronType(*child);
            CppBlockPtr b2 = child_s->getResetSoftBlock(key);
            block_if_data->appendBlock(b2);
        }
        return block_if_data;
    }

    CppBlockPtr ObjectClass::getResetHardBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = std::make_shared<CppBlock>();
        if (type.getExtends() != nullptr)
        {
            const auto extends_s = FromAronType(*type.getExtends());
            block_if_data->addLine(extends_s->getFullCppTypename() + "::resetHard();");
        }

        for (const auto& [key, child] : type.getMemberTypes())
        {
            const auto child_s = FromAronType(*child);
            CppBlockPtr b2 = child_s->getResetHardBlock(key);
            block_if_data->appendBlock(b2);
        }
        return block_if_data;
    }

    CppBlockPtr ObjectClass::getWriteTypeBlock(const std::string&, const std::string&, std::string&) const
    {
        static const std::string OBJECT_MEMBERS_ACCESSOR = ARON_VARIABLE_PREFIX + "_objectMembers";
        static const std::string OBJECT_EXTENDS_ACCESSOR = ARON_VARIABLE_PREFIX + "_objectExtends";

        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("std::map<std::string, T> " + OBJECT_MEMBERS_ACCESSOR + ";");

        if (type.getExtends() != nullptr)
        {
            const auto extends_s = FromAronType(*type.getExtends());
            b->addLine("// get base class of " + this->getFullCppTypename());
            b->addLine("auto " + OBJECT_EXTENDS_ACCESSOR + " = " + extends_s->getFullCppTypename() + "::writeType(" + ARON_WRITER_ACCESSOR + ", ::armarx::aron::type::Maybe::eNone);");
        }

        b->addLine("// members of " + this->getFullCppTypename());
        for (const auto& [key, child] : type.getMemberTypes())
        {
            const auto child_s = FromAronType(*child);
            std::string child_return_variant;
            CppBlockPtr child_b = child_s->getWriteTypeBlock(child_s->getFullCppTypename(), key, child_return_variant);
            b->appendBlock(child_b);
            b->addLine(OBJECT_MEMBERS_ACCESSOR + ".emplace(\"" + key + "\", " + child_return_variant + ");");
        }

        if (type.getExtends() != nullptr)
        {
            b->addLine("return " + ARON_WRITER_ACCESSOR + ".writeObject(\"" + type.getObjectName() + "\", "+ARON_MAYBE_TYPE_ACCESSOR+", " + OBJECT_MEMBERS_ACCESSOR + ", " + OBJECT_EXTENDS_ACCESSOR + ");");
        }
        else
        {
            b->addLine("return " + ARON_WRITER_ACCESSOR + ".writeObject(\"" + type.getObjectName() + "\", "+ARON_MAYBE_TYPE_ACCESSOR+", " + OBJECT_MEMBERS_ACCESSOR + ", std::nullopt);");
        }

        return b;
    }

    CppBlockPtr ObjectClass::getWriteBlock(const std::string&, std::string&) const
    {
        static const std::string OBJECT_MEMBERS_ACCESSOR = ARON_VARIABLE_PREFIX + "_objectMembers";
        static const std::string OBJECT_EXTENDS_ACCESSOR = ARON_VARIABLE_PREFIX + "_objectExtends";

        CppBlockPtr block_if_data = std::make_shared<CppBlock>();

        block_if_data->addLine("std::map<std::string, T> " + OBJECT_MEMBERS_ACCESSOR + ";");

        block_if_data->addLine("std::optional<T> " + OBJECT_EXTENDS_ACCESSOR + ";");
        if (type.getExtends() != nullptr)
        {
            const auto extends_s = FromAronType(*type.getExtends());
            block_if_data->addLine("// write base class of " + this->getFullCppTypename());
            block_if_data->addLine(OBJECT_EXTENDS_ACCESSOR + " = " + extends_s->getFullCppTypename() + "::write(" + ARON_WRITER_ACCESSOR + ");");
        }

        block_if_data->addLine("// members of " + this->getFullCppTypename());
        for (const auto& [key, child] : type.getMemberTypes())
        {
            const auto child_s = FromAronType(*child);
            std::string child_return_variant;
            CppBlockPtr child_b = child_s->getWriteBlock(key, child_return_variant);
            block_if_data->addLine("auto " + child_return_variant + " = " + ARON_WRITER_ACCESSOR + ".writeNull();");
            block_if_data->appendBlock(child_b);
            block_if_data->addLine(OBJECT_MEMBERS_ACCESSOR + ".emplace(\"" + key + "\", " + child_return_variant + ");");
        }

        block_if_data->addLine("return " + ARON_WRITER_ACCESSOR + ".writeDict("+OBJECT_MEMBERS_ACCESSOR+", " + OBJECT_EXTENDS_ACCESSOR + "); // of top level object " + getCoreCppTypename());
        return block_if_data;
    }

    CppBlockPtr ObjectClass::getReadBlock(const std::string&, const std::string& variantAccessor) const
    {
        static const std::string OBJECT_MEMBERS_ACCESSOR = ARON_VARIABLE_PREFIX + "_objectMembers";
        static const std::string OBJECT_EXTENDS_ACCESSOR = ARON_VARIABLE_PREFIX + "_objectExtends";

        CppBlockPtr block_if_data = std::make_shared<CppBlock>();
        block_if_data->addLine("std::map<std::string, TNonConst> " + OBJECT_MEMBERS_ACCESSOR + ";");

        if (type.getExtends() != nullptr)
        {
            const auto extends_s = FromAronType(*type.getExtends());
            block_if_data->addLine(extends_s->getFullCppTypename() + "::read(" + ARON_READER_ACCESSOR + ", "+variantAccessor+");");
        }

        block_if_data->addLine("" + ARON_READER_ACCESSOR + ".readDict("+variantAccessor+", "+OBJECT_MEMBERS_ACCESSOR+"); // of top level object " + getCoreCppTypename());

        for (const auto& [key, child] : type.getMemberTypes())
        {
            const auto child_s = FromAronType(*child);
            const std::string child_variant_accessor = OBJECT_MEMBERS_ACCESSOR + ".at(\"" + key + "\")";
            block_if_data->appendBlock(child_s->getReadBlock(key, child_variant_accessor));
        }
        return block_if_data;
    }

    CppBlockPtr ObjectClass::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
    {
        CppBlockPtr block_if_data = std::make_shared<CppBlock>();
        if (type.getExtends() != nullptr)
        {
            const auto extends_s = FromAronType(*type.getExtends());
            block_if_data->addLine("if (not (" + extends_s->getFullCppTypename() + "::operator== (" + otherInstanceAccessor + ")))");
            block_if_data->addLineAsBlock("return false;");
        }
        for (const auto& [key, child] : type.getMemberTypes())
        {
            auto child_s = FromAronType(*child);
            CppBlockPtr b2 = child_s->getEqualsBlock(key, otherInstanceAccessor + "." + key);
            block_if_data->appendBlock(b2);
        }
        return block_if_data;
    }
}

