/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include "../visitor/Visitor.h"
#include "../rw/Reader.h"
#include "../rw/Writer.h"

namespace armarx::aron::data
{
    // prototype
    template <class ConverterImplementation>
    typename ConverterImplementation::WriterReturnType readAndWrite(typename ConverterImplementation::ReaderInputType& o);

    /// Converter struct providing the needed methods.
    /// WriterImplementation is a writer class, TODO: add concepts
    template <class ReaderImplementation, class WriterImplementation, class DerivedT>
    struct Converter :
            virtual public Visitor<typename ReaderImplementation::InputType>
    {
        using WriterReturnType = typename WriterImplementation::ReturnType;
        using ReaderInputType = typename ReaderImplementation::InputType;
        using ReaderInputTypeNonConst = typename ReaderImplementation::InputTypeNonConst;

        ReaderImplementation r;
        WriterImplementation w;
        WriterReturnType last_returned;

        virtual ~Converter() = default;

        virtual void visitDict(ReaderInputType& o) override
        {
            std::map<std::string, ReaderInputTypeNonConst> elementsOfInput;
            std::map<std::string, WriterReturnType> elementsReturn;
            r.readDict(o, elementsOfInput);
            for (const auto& [key, value] : elementsOfInput)
            {
                auto converted = readAndWrite<DerivedT>(value);
                elementsReturn.insert({key, converted});
            }

            last_returned = w.writeDict(elementsReturn);
        };
        virtual void visitList(ReaderInputType& o) override
        {
            std::vector<ReaderInputTypeNonConst> elementsOfInput;
            std::vector<WriterReturnType> elementsReturn;
            r.readList(o, elementsOfInput);
            for (const auto& value : elementsOfInput)
            {
                auto converted = readAndWrite<DerivedT>(value);
                elementsReturn.push_back(converted);
            }
            last_returned = w.writeList(elementsReturn);
        };
        virtual void visitNDArray(ReaderInputType& o) override
        {
            std::string type;
            std::vector<int> shape;
            std::vector<unsigned char> data;
            r.readNDArray(o, shape, type, data);
            last_returned = w.writeNDArray(shape, type, data.data());
        };
        virtual void visitInt(ReaderInputType& o) override
        {
            int i;
            r.readInt(o, i);
            last_returned = w.writeInt(i);
        };
        virtual void visitLong(ReaderInputType& o) override
        {
            long i;
            r.readLong(o, i);
            last_returned = w.writeLong(i);
        };
        virtual void visitFloat(ReaderInputType& o) override
        {
            float i;
            r.readFloat(o, i);
            last_returned = w.writeFloat(i);
        };
        virtual void visitDouble(ReaderInputType& o) override
        {
            double i;
            r.readDouble(o, i);
            last_returned = w.writeDouble(i);
        };
        virtual void visitBool(ReaderInputType& o) override
        {
            bool i;
            r.readBool(o, i);
            last_returned = w.writeBool(i);
        };
        virtual void visitString(ReaderInputType& o) override
        {
            std::string i;
            r.readString(o, i);
            last_returned = w.writeString(i);
        };
    };

    /// the function to read from a variant and write to a writer T
    /// returns the returntype of T
    template <class ConverterImplementation>
    typename ConverterImplementation::WriterReturnType readAndWrite(typename ConverterImplementation::ReaderInputType& o)
    {
        ConverterImplementation v;
        data::visit(v, o);
        return v.last_returned;
    }
}
