/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <map>
#include <string>
#include <vector>

#include "../RecursiveVisitor.h"
#include "../../variant/All.h"

namespace armarx::aron::data
{
    /**
     * @see type/visitor/variant/VariantVisitor.h
     */
    struct ConstVariantVisitor : virtual public Visitor<const data::VariantPtr>
    {
        virtual data::Descriptor getDescriptor(Input& n) override
        {
            if (!n)
            {
                return data::Descriptor::eUnknown;
            }
            return n->getDescriptor();
        }

        virtual ~ConstVariantVisitor() = default;
    };

    /**
     * @see type/visitor/variant/VariantVisitor.h
     */
    struct ConstTypedVariantVisitor : virtual public TypedVisitor<const data::VariantPtr, const type::VariantPtr>
    {
        virtual type::Descriptor getDescriptor(TypeInput& n) override
        {
            if (!n)
            {
                return type::Descriptor::eUnknown;
            }
            return n->getDescriptor();
        }

        virtual ~ConstTypedVariantVisitor() = default;
    };

    /**
     * @see type/visitor/variant/VariantVisitor.h
     */
    struct RecursiveConstVariantVisitor : virtual public RecursiveVisitor<const data::VariantPtr>
    {
        virtual data::Descriptor getDescriptor(Input& n) override
        {
            if (!n)
            {
                return data::Descriptor::eUnknown;
            }
            return n->getDescriptor();
        }

        virtual std::map<std::string, InputNonConst> getDictElements(Input& n) override
        {
            auto x = data::Dict::DynamicCastAndCheck(n);
            return x->getElements();
        }

        virtual std::vector<InputNonConst> getListElements(Input& n) override
        {
            auto x = data::List::DynamicCastAndCheck(n);
            return x->getElements();
        }

        virtual ~RecursiveConstVariantVisitor() = default;
    };

    /**
     * @see type/visitor/variant/VariantVisitor.h
     */
    struct RecursiveVariantVisitor : virtual public RecursiveVisitor<data::VariantPtr>
    {
        virtual data::Descriptor getDescriptor(Input& n) override
        {
            if (!n)
            {
                return data::Descriptor::eUnknown;
            }
            return n->getDescriptor();
        }

        virtual std::map<std::string, InputNonConst> getDictElements(Input& n) override
        {
            auto x = data::Dict::DynamicCastAndCheck(n);
            return x->getElements();
        }

        virtual std::vector<InputNonConst> getListElements(Input& n) override
        {
            auto x = data::List::DynamicCastAndCheck(n);
            return x->getElements();
        }

        virtual ~RecursiveVariantVisitor() = default;
    };

    /**
     * @see type/visitor/variant/VariantVisitor.h
     */
    struct RecursiveConstTypedVariantVisitor : virtual public RecursiveTypedVisitor<const data::VariantPtr, const type::VariantPtr>
    {
        virtual type::Descriptor getDescriptor(TypeInput& n) override
        {
            if (!n)
            {
                return type::Descriptor::eUnknown;
            }
            return n->getDescriptor();
        }

        virtual std::map<std::string, std::pair<DataInputNonConst, TypeInputNonConst>> getObjectElements(DataInput& o, TypeInput& t) override
        {
            std::map<std::string, std::pair<data::VariantPtr, type::VariantPtr>> ret;
            auto x = data::Dict::DynamicCastAndCheck(o);
            auto y = type::Object::DynamicCastAndCheck(t);
            for (const auto& [key, e] : x->getElements())
            {
                auto ct = y->getMemberType(key);
                ret.insert({key, {e, ct}});
            }
            return ret;
        }
        virtual std::map<std::string, std::pair<DataInputNonConst, TypeInputNonConst>> getDictElements(DataInput& o, TypeInput& t) override
        {
            std::map<std::string, std::pair<data::VariantPtr, type::VariantPtr>> ret;
            auto x = data::Dict::DynamicCastAndCheck(o);
            auto y = type::Dict::DynamicCastAndCheck(t);
            auto ac = y->getAcceptedType();

            for (const auto& [key, e] : x->getElements())
            {
                ret.insert({key, {e, ac}});
            }
            return ret;
        }
        virtual std::vector<std::pair<DataInputNonConst, TypeInputNonConst>> getListElements(DataInput& o, TypeInput& t) override
        {
            std::vector<std::pair<data::VariantPtr, type::VariantPtr>> ret;
            auto x = data::List::DynamicCastAndCheck(o);
            auto y = type::List::DynamicCastAndCheck(t);
            auto ac = y->getAcceptedType();

            for (const auto& e : x->getElements())
            {
                ret.push_back({e, ac});
            }
            return ret;
        }
        virtual std::pair<std::pair<DataInputNonConst, TypeInputNonConst>, std::pair<DataInputNonConst, TypeInputNonConst>>
        getPairElements(DataInput& o, TypeInput& t) override
        {
            auto x = data::List::DynamicCastAndCheck(o);
            auto y = type::Pair::DynamicCastAndCheck(t);

            auto e0 = x->getElement(0);
            auto ac0 = y->getFirstAcceptedType();
            auto e1 = x->getElement(1);
            auto ac1 = y->getSecondAcceptedType();
            return {{e0, ac0}, {e1, ac1}};
        }
        virtual std::vector<std::pair<DataInputNonConst, TypeInputNonConst>> getTupleElements(DataInput& o, TypeInput& t) override
        {
            std::vector<std::pair<data::VariantPtr, type::VariantPtr>> ret;
            auto x = data::List::DynamicCastAndCheck(o);
            auto y = type::Tuple::DynamicCastAndCheck(t);

            unsigned int i = 0;
            for (const auto& e : x->getElements())
            {
                auto ac = y->getAcceptedType(i++);
                ret.push_back({e, ac});
            }
            return ret;
        }

        virtual ~RecursiveConstTypedVariantVisitor() = default;
    };
}
