#pragma once

#include <string>

#include <RobotAPI/libraries/aron/core/data/variant/All.h>


namespace armarx::aron::visitor
{

    /**
     * @brief A visitor traversing an `aron::data::dto::GenericDataPtr` using
     * `aron::datanavigator::Navigator`s, taking care of type checking,
     * casting and hierarchical traversal.
     *
     * To use it, create a class deriving from `AronDataVisitor` and
     * override the overloads of `visitEnter()`, `visitExit()` and `visit()`
     * for the types you would like to handle (see blow for a copy-and-paste
     * example).
     * Then, instantiate your derived visitor and call `applyTo()`, passing
     * the `AronDataPtr` or the root `Navigator`.
     *
     * For dicts and lists, you will get a `visitEnter()`/`visitExit()` pair.
     * For types without children (e.g. primitive types), you get a single
     * `visit()` call.
     *
     * Each visit function should return true to continue traversal or
     * false to stop it. All default implementations return true.
     *
     * To the the location of the passed navigator in the original `AronData`,
     * use `navigator.getPath()`.
     *
     * @see  `aron::data::dto::GenericDataPtr`
     * @see  `aron::datanavigator::Navigator`
     * @see `std::stack` To manage data with stack semantics.
     */
    class DataVisitor
    {
    public:

        // Define types to ease writing code in this and derived classes.

        using DataNavigator = data::Variant;

        using DictDataNavigator = data::Dict;
        using ListDataNavigator = data::List;

        using BoolDataNavigator = data::Bool;
        using DoubleDataNavigator = data::Double;
        using FloatDataNavigator = data::Float;
        using IntDataNavigator = data::Int;
        using LongDataNavigator = data::Long;
        using NDArrayDataNavigator = data::NDArray;
        using StringDataNavigator = data::String;


    public:

        DataVisitor();
        virtual ~DataVisitor();


        bool applyTo(data::dto::GenericDataPtr& aronData);
        bool applyTo(DataNavigator& navigator);
        bool applyTo(const std::string& key, DataNavigator& navigator);



        virtual bool visitEnter(DictDataNavigator& dict)
        {
            (void) dict;
            return true;
        }
        virtual bool visitExit(DictDataNavigator& dict)
        {
            (void) dict;
            return true;
        }
        virtual bool visitEnter(ListDataNavigator& list)
        {
            (void) list;
            return true;
        }
        virtual bool visitExit(ListDataNavigator& list)
        {
            (void) list;
            return true;
        }

        virtual bool visit(BoolDataNavigator& b)
        {
            (void) b;
            return true;
        }
        virtual bool visit(DoubleDataNavigator& d)
        {
            (void) d;
            return true;
        }
        virtual bool visit(FloatDataNavigator& f)
        {
            (void) f;
            return true;
        }
        virtual bool visit(IntDataNavigator& i)
        {
            (void) i;
            return true;
        }
        virtual bool visit(LongDataNavigator& l)
        {
            (void) l;
            return true;
        }
        virtual bool visit(NDArrayDataNavigator& array)
        {
            (void) array;
            return true;
        }
        virtual bool visit(StringDataNavigator& string)
        {
            (void) string;
            return true;
        }


        virtual bool visitEnter(const std::string& key, DictDataNavigator& dict)
        {
            (void) key;
            return visitEnter(dict);
        }
        virtual bool visitExit(const std::string& key, DictDataNavigator& dict)
        {
            (void) key;
            return visitExit(dict);
        }
        virtual bool visitEnter(const std::string& key, ListDataNavigator& list)
        {
            (void) key;
            return visitEnter(list);
        }
        virtual bool visitExit(const std::string& key, ListDataNavigator& list)
        {
            (void) key;
            return visitExit(list);
        }

        virtual bool visit(const std::string& key, BoolDataNavigator& b)
        {
            (void) key;
            return visit(b);
        }
        virtual bool visit(const std::string& key, DoubleDataNavigator& d)
        {
            (void) key;
            return visit(d);
        }
        virtual bool visit(const std::string& key, FloatDataNavigator& f)
        {
            (void) key;
            return visit(f);
        }
        virtual bool visit(const std::string& key, IntDataNavigator& i)
        {
            (void) key;
            return visit(i);
        }
        virtual bool visit(const std::string& key, LongDataNavigator& l)
        {
            (void) key;
            return visit(l);
        }
        virtual bool visit(const std::string& key, NDArrayDataNavigator& array)
        {
            (void) key;
            return visit(array);
        }
        virtual bool visit(const std::string& key, StringDataNavigator& string)
        {
            (void) key;
            return visit(string);
        }


    private:

        bool applyToChildren(DictDataNavigator& navigator);
        bool applyToChildren(ListDataNavigator& navigator);

    };

}


/* Copy-and-paste example below.
 * Remove functions you don't need.
 * Add `const std::string& key, ` to parameter list if you need an items's key
 * in the parent container.
 */

#if 0

struct MyDerivedAronDataVisitor : public armarx::aron::visitor::AronDataVisitor
{

    bool visitEnter(DictDataNavigator& dict) override
    {
        (void) dict;
        return true;
    }
    bool visitExit(DictDataNavigator& dict) override
    {
        (void) dict;
        return true;
    }
    bool visitEnter(ListDataNavigator& list) override
    {
        (void) list;
        return true;
    }
    bool visitExit(ListDataNavigator& list) override
    {
        (void) list;
        return true;
    }

    bool visit(BoolDataNavigator& b) override
    {
        (void) b;
        return true;
    }
    bool visit(DoubleDataNavigator& d) override
    {
        (void) d;
        return true;
    }
    bool visit(FloatDataNavigator& f) override
    {
        (void) f;
        return true;
    }
    bool visit(IntDataNavigator& i) override
    {
        (void) i;
        return true;
    }
    bool visit(LongDataNavigator& l) override
    {
        (void) l;
        return true;
    }
    bool visit(NDArrayDataNavigator& array) override
    {
        (void) array;
        return true;
    }
    bool visit(StringDataNavigator& string) override
    {
        (void) string;
        return true;
    }

};

#endif
