/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <map>
#include <memory>
#include <numeric>
#include <set>

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/data/variant/All.h>
#include <RobotAPI/libraries/aron/core/type/variant/All.h>

#include <RobotAPI/libraries/aron/core/type/visitor/variant/VariantVisitor.h>
#include <RobotAPI/libraries/aron/core/data/visitor/variant/VariantVisitor.h>

#include <RobotAPI/libraries/aron/converter/json/NLohmannJSONConverter.h>

namespace armarx::aron
{
    namespace type::matrix
    {
        const std::vector<type::matrix::ElementType> AllTypes = {
            type::matrix::ElementType::int16,
            type::matrix::ElementType::int32,
            type::matrix::ElementType::int64,
            type::matrix::ElementType::float32,
            type::matrix::ElementType::float64
        };
    }

    namespace type::quaternion
    {
        const std::vector<type::quaternion::ElementType> AllTypes = {
            type::quaternion::float32,
            type::quaternion::float64
        };
    }

    namespace type::image
    {
        const std::vector<type::image::PixelType> AllTypes = {
            type::image::PixelType::rgb24,
            type::image::PixelType::depth32
        };
    }

    namespace type::pointcloud
    {
        const std::vector<type::pointcloud::VoxelType> AllTypes = {
            type::pointcloud::VoxelType::PointXYZ,
            type::pointcloud::VoxelType::PointXYZHSV,
            type::pointcloud::VoxelType::PointXYZI,
            type::pointcloud::VoxelType::PointXYZL,
            type::pointcloud::VoxelType::PointXYZRGB,
            type::pointcloud::VoxelType::PointXYZRGBA,
            type::pointcloud::VoxelType::PointXYZRGBL
        };
    }

    namespace util
    {
        // generate i in [min, max)
        int generateRandom(int max, int min)
        {
            if (max < min)
            {
                throw error::AronException(__PRETTY_FUNCTION__, "The max value "+std::to_string(max)+" is lower than the min value "+std::to_string(min));
            }
            int random = (std::rand() % (max - min)) + min;
            return random;
        }

        template <class T>
        T getRandomElement(const std::vector<T>& vec)
        {
            int i = generateRandom(vec.size(), 0);
            return vec.at(i);
        }

        template <class ValueT>
        std::string getRandomKey(const std::map<std::string, ValueT>& m)
        {
            std::vector<std::string> keys;
            for (const auto [k, _] : m)
            {
                keys.push_back(k);
            }
            return getRandomElement(keys);
        }

        std::string getRandomKey(const std::set<std::string>& set)
        {
            std::vector<std::string> keys;
            for (const auto& k : set)
            {
                keys.push_back(k);
            }
            return getRandomElement(keys);
        }

        bool fiftyPercentChance()
        {
            return generateRandom(2, 0);
        }

        std::string generateRandomWord(const std::set<std::string>& usedKeys = {})
        {
            // https://randomwordgenerator.com/
            // script to convert random words into string:
            // jQuery("#result li i").remove(); var str = ""; jQuery("#result li").each(function() {str += jQuery(this).html() + ", " }); console.log(str);
            const std::vector<std::string> words =
            {
                "jaw", "river", "bow", "profession", "heat", "interference", "slave", "population", "deport", "redeem", "digress", "penny", "cousin", "beef", "Bible", "fuss",
                "urgency", "tasty", "innovation", "upset", "gold", "day", "remunerate", "strain", "district", "panel", "choke", "rack", "parameter", "despair", "extort", "country",
                "hesitate", "record", "brand", "confusion", "discreet", "accept", "lifestyle", "option", "corn", "charity", "miss", "viable", "glance", "norm", "meet", "bird",
                "ribbon", "guideline"
            };

            int i = generateRandom(words.size(), 0);
            std::string key = words.at(i);

            while (usedKeys.count(key) > 0)
            {
                key = generateRandomWord(usedKeys);
            }

            return key;
        }

        std::vector<unsigned char> generateRandomBlob(unsigned int size)
        {
            std::vector<unsigned char> new_blob(size, 0);
            for (unsigned int i = 0; i < size; ++i)
            {
                new_blob[i] = (generateRandom(128, 0));
            }
            return new_blob;
        }

        struct RandomTypeGenerator : public type::Visitor<const type::Descriptor>
        {
            type::VariantPtr type;

            static type::VariantPtr GenerateRandomType(bool can_be_maybe = true, bool must_be_object = false)
            {
                RandomTypeGenerator nt;
                if (must_be_object)
                {
                    type::visit(nt, type::Descriptor::eObject);
                }
                else
                {
                    type::visit(nt, util::getRandomElement(type::AllDescriptors));
                }

                if (can_be_maybe)
                {
                    type::Maybe nextMaybeType = util::getRandomElement(type::AllMaybeTypes);
                    nt.type->setMaybe(nextMaybeType);
                }
                return nt.type;
            }

            virtual type::Descriptor getDescriptor(Input& t) override { return t; }
            virtual void visitObject(Input&) override
            {
                std::string objectName = util::generateRandomWord();

                int members = util::generateRandom(4, 2);
                std::set<std::string> usedKeys = {objectName};
                std::map<std::string, type::VariantPtr> elements;
                for (int i = 0; i < members; ++i)
                {
                    std::string key = util::generateRandomWord(usedKeys);
                    usedKeys.insert(key);
                    elements.insert({key, GenerateRandomType()});
                }
                type = std::make_shared<type::Object>(objectName, elements);
            }
            virtual void visitDict(Input&) override { type = std::make_shared<type::Dict>(GenerateRandomType()); }
            virtual void visitPair(Input&) override { type = std::make_shared<type::Pair>(GenerateRandomType(), GenerateRandomType()); }
            virtual void visitTuple(Input&) override
            {
                int members = util::generateRandom(4, 2);
                std::vector<type::VariantPtr> acceptedTypes;
                for (int i = 0; i < members; ++i)
                {
                    acceptedTypes.push_back(GenerateRandomType());
                }

                type = std::make_shared<type::Tuple>(acceptedTypes);
            }
            virtual void visitList(Input&) override { type = std::make_shared<type::List>(GenerateRandomType()); }
            virtual void visitMatrix(Input&) override
            {
                auto x = std::make_shared<type::Matrix>();
                auto t = util::getRandomElement(type::matrix::AllTypes);
                x->setElementType(t);
                x->setRows(util::generateRandom(10, 1));
                x->setRows(util::generateRandom(10, 1));
                type = x;
            }
            virtual void visitNDArray(Input& t) override { visitMatrix(t); }
            virtual void visitQuaternion(Input&) override
            {
                auto x = std::make_shared<type::Quaternion>();
                auto t = util::getRandomElement(type::quaternion::AllTypes);
                x->setElementType(t);
                type = x;
            }
            virtual void visitOrientation(Input&) override { type = std::make_shared<type::Orientation>(); }
            virtual void visitPosition(Input&) override { type = std::make_shared<type::Position>(); }
            virtual void visitPose(Input&) override { type = std::make_shared<type::Pose>(); }
            virtual void visitImage(Input&) override
            {
                auto t = std::make_shared<type::Image>();
                auto x = util::getRandomElement(type::image::AllTypes);
                t->setPixelType(x);
                type = t;
            }
            virtual void visitPointCloud(Input&) override
            {
                auto t = std::make_shared<type::PointCloud>();
                auto x = util::getRandomElement(type::pointcloud::AllTypes);
                t->setVoxelType(x);
                type = t;
            }
            virtual void visitIntEnum(Input&) override
            {
                std::string intEnumName = util::generateRandomWord();

                std::set<std::string> usedKeys = {intEnumName};
                std::map<std::string, int> acceptedValues;
                for (int i = 0; i < util::generateRandom(5, 2); ++i)
                {
                    std::string key = util::generateRandomWord(usedKeys);
                    usedKeys.insert(key);

                    acceptedValues.insert({key, i});
                }
                type = std::make_shared<type::IntEnum>(intEnumName, acceptedValues);
            }
            virtual void visitInt(Input&) override { type = std::make_shared<type::Int>(); }
            virtual void visitLong(Input&) override { type = std::make_shared<type::Long>(); }
            virtual void visitFloat(Input&) override { type = std::make_shared<type::Float>(); }
            virtual void visitDouble(Input&) override { type = std::make_shared<type::Double>(); }
            virtual void visitBool(Input&) override { type = std::make_shared<type::Bool>(); }
            virtual void visitString(Input&) override { type = std::make_shared<type::String>(); }
            virtual void visitTime(Input&) override { type = std::make_shared<type::Time>(); }
            virtual void visitUnknown(Input&) override { type = GenerateRandomType(); }
        };

        struct EmptyDataFromTypeGenerator : public type::Visitor<const type::Descriptor>
        {
            type::VariantPtr type;
            data::VariantPtr data;

            static data::VariantPtr GenerateRandomFromType(const type::VariantPtr& t)
            {
                EmptyDataFromTypeGenerator nt;
                nt.type = t;
                type::visit(nt, util::getRandomElement(type::AllDescriptors));
                return nt.data;
            }

            virtual type::Descriptor getDescriptor(Input& t) override { return t; }
            virtual void visitObject(Input&) override
            {
                auto da = std::make_shared<data::Dict>();

                type::ObjectPtr t = type::Object::DynamicCastAndCheck(type);
                auto d = std::make_shared<data::Dict>();
                for (const auto& [k, tt] : t->getMemberTypes())
                {
                    d->addElement(k, GenerateRandomFromType(tt));
                }
                data = da;
            }
            virtual void visitDict(Input&) override { data = std::make_shared<data::Dict>(); }
            virtual void visitPair(Input&) override { data = std::make_shared<data::List>(); }
            virtual void visitTuple(Input&) override { data = std::make_shared<data::List>(); }
            virtual void visitList(Input&) override { data = std::make_shared<data::List>(); }
            virtual void visitMatrix(Input&) override { data = std::make_shared<data::NDArray>(); }
            virtual void visitNDArray(Input& t) override { data = std::make_shared<data::NDArray>(); }
            virtual void visitQuaternion(Input&) override { data = std::make_shared<data::NDArray>(); }
            virtual void visitOrientation(Input&) override { data = std::make_shared<data::NDArray>(); }
            virtual void visitPosition(Input&) override { data = std::make_shared<data::NDArray>(); }
            virtual void visitPose(Input&) override { data = std::make_shared<data::NDArray>(); }
            virtual void visitImage(Input&) override { data = std::make_shared<data::NDArray>(); }
            virtual void visitPointCloud(Input&) override { data = std::make_shared<data::NDArray>(); }
            virtual void visitIntEnum(Input&) override { data = std::make_shared<data::Int>(); }
            virtual void visitInt(Input&) override { data = std::make_shared<data::Int>(); }
            virtual void visitLong(Input&) override { data = std::make_shared<data::Long>(); }
            virtual void visitFloat(Input&) override { data = std::make_shared<data::Float>(); }
            virtual void visitDouble(Input&) override { data = std::make_shared<data::Double>(); }
            virtual void visitBool(Input&) override { data = std::make_shared<data::Bool>(); }
            virtual void visitString(Input&) override { data = std::make_shared<data::String>(); }
            virtual void visitTime(Input&) override { data = std::make_shared<data::Long>(); }
            virtual void visitUnknown(Input&) override { data = GenerateRandomFromType(type); }
        };
    }

    class Randomizer
    {
    public:
        Randomizer()
        {
            std::srand(std::time(nullptr));
        };

        type::VariantPtr generateRandomObjectType() const
        {
            return util::RandomTypeGenerator::GenerateRandomType(false, true);
        }

        data::VariantPtr generateEmptyAronDataFromType(const type::VariantPtr& type, bool ignore_maybe = false) const
        {
            if (type->getMaybe() != type::Maybe::eNone && !ignore_maybe)
            {
                if (util::fiftyPercentChance())
                {
                    return nullptr;
                }
            }

            auto data = util::EmptyDataFromTypeGenerator::GenerateRandomFromType(type);
            return data;
        }

        void initializeRandomly(data::VariantPtr& data, const type::VariantPtr& type) const
        {
            data = generateEmptyAronDataFromType(type);
            if (!data)
            {
                if (type->getMaybe() == type::Maybe::eNone)
                {
                    throw error::AronException(__PRETTY_FUNCTION__, "The data is null but the type is not a maybe type. This is not valid.");
                }
                return;
            }

            // Containers
            type::Descriptor desc = type->getDescriptor();
            switch (desc)
            {
                case type::Descriptor::eObject:
                {
                    auto t = type::Object::DynamicCastAndCheck(type);
                    auto d = data::Dict::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eList:
                {
                    auto t = type::List::DynamicCastAndCheck(type);
                    auto d = data::List::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eDict:
                {
                    auto t = type::Dict::DynamicCastAndCheck(type);
                    auto d = data::Dict::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eTuple:
                {
                    auto t = type::Tuple::DynamicCastAndCheck(type);
                    auto d = data::List::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::ePair:
                {
                    auto t = type::Pair::DynamicCastAndCheck(type);
                    auto d = data::List::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eIntEnum:
                {
                    auto t = type::IntEnum::DynamicCastAndCheck(type);
                    auto d = data::Int::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eMatrix:
                {
                    auto t = type::Matrix::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eNDArray:
                {
                    auto t = type::NDArray::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eOrientation:
                {
                    auto t = type::Orientation::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eQuaternion:
                {
                    auto t = type::Quaternion::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eImage:
                {
                    auto t = type::Image::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::ePointCloud:
                {
                    auto t = type::PointCloud::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::ePosition:
                {
                    auto t = type::Position::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::ePose:
                {
                    auto t = type::Pose::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eInt:
                {
                    auto t = type::Int::DynamicCastAndCheck(type);
                    auto d = data::Int::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eLong:
                {
                    auto t = type::Long::DynamicCastAndCheck(type);
                    auto d = data::Long::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eFloat:
                {
                    auto t = type::Float::DynamicCastAndCheck(type);
                    auto d = data::Float::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eDouble:
                {
                    auto t = type::Double::DynamicCastAndCheck(type);
                    auto d = data::Double::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eString:
                {
                    auto t = type::String::DynamicCastAndCheck(type);
                    auto d = data::String::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eBool:
                {
                    auto t = type::Bool::DynamicCastAndCheck(type);
                    auto d = data::Bool::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eTime:
                {
                    auto t = type::Time::DynamicCastAndCheck(type);
                    auto d = data::Long::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eUnknown:
                {
                    throw error::AronException(__PRETTY_FUNCTION__, "No valid type found!");
                }
            }

        }

        void initializeRandomly(data::DictPtr& data, const type::Object& type) const
        {
            for (auto& [key, nextData] : data->getElements())
            {
                initializeRandomly(nextData, type.getMemberTypes()[key]);
                data->removeElement(key);
                data->addElement(key, nextData);
            }
        }

        void initializeRandomly(data::DictPtr& data, const type::Dict& type) const
        {
            data->clear();
            int numElements = util::generateRandom(5, 1);
            std::set<std::string> usedKeys;
            for (int i = 0; i < numElements; ++i)
            {
                std::string key = util::generateRandomWord(usedKeys);
                usedKeys.insert(key);
                data::VariantPtr newData = generateEmptyAronDataFromType(type.getAcceptedType());
                initializeRandomly(newData, type.getAcceptedType());
                data->addElement(key, newData);
            }
        }

        void initializeRandomly(data::ListPtr& data, const type::Tuple& type) const
        {
            std::vector<data::VariantPtr> newElements;

            unsigned int i = 0;
            for (auto& nextData : data->getElements())
            {
                initializeRandomly(nextData, type.getAcceptedType(i++));
                newElements.push_back(nextData);
            }

            data->clear();
            for (const auto& el : newElements)
            {
                data->addElement(el);
            }
        }

        void initializeRandomly(data::ListPtr& data, const type::Pair& type) const
        {
            data::VariantPtr first = data->getElement(0);
            data::VariantPtr second = data->getElement(1);
            initializeRandomly(first, type.getFirstAcceptedType());
            initializeRandomly(second, type.getSecondAcceptedType());

            data->clear();
            data->addElement(first);
            data->addElement(second);
        }

        void initializeRandomly(data::ListPtr& data, const type::List& type) const
        {
            data->clear();
            int numElements = util::generateRandom(5, 1);
            for (int i = 0; i < numElements; ++i)
            {
                data::VariantPtr newData = generateEmptyAronDataFromType(type.getAcceptedType());
                initializeRandomly(newData, type.getAcceptedType());
                data->addElement(newData);
            }
        }

        void initializeRandomly(data::NDArrayPtr& data, const type::Variant& type) const
        {
            std::vector<int> dims = data->getShape();
            int bytes = std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>());
            std::vector<unsigned char> blob = util::generateRandomBlob(bytes);
            data->setData(bytes, blob.data());
        }

        void initializeRandomly(data::IntPtr& data, const type::Variant& type) const
        {
            data->setValue(util::generateRandom(1000, -1000));
        }

        void initializeRandomly(data::LongPtr& data, const type::Variant& type) const
        {
            data->setValue(util::generateRandom(1000, -1000));
        }

        void initializeRandomly(data::FloatPtr& data, const type::Variant& type) const
        {
            data->setValue(util::generateRandom(1000, -1000));
        }

        void initializeRandomly(data::DoublePtr& data, const type::Variant& type) const
        {
            data->setValue(util::generateRandom(1000, -1000));
        }

        void initializeRandomly(data::BoolPtr& data, const type::Variant& type) const
        {
            data->setValue(util::fiftyPercentChance());
        }

        void initializeRandomly(data::StringPtr& data, const type::Variant& type) const
        {
            data->setValue(util::generateRandomWord());
        }
    };
}

