/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#include "VariantWriter.h"

#include <RobotAPI/libraries/aron/core/type/variant/All.h>

#include <memory>
#include <numeric>


namespace armarx::aron::type::writer
{
    // interface
    aron::type::VariantPtr VariantWriter::writeObject(const std::string& name, const type::Maybe maybe, const std::map<std::string, aron::type::VariantPtr>& memberTypes, const std::optional<aron::type::VariantPtr>& extends)
    {
        auto o = std::make_shared<type::Object>(name, memberTypes);
        o->setMaybe(maybe);
        if (extends.has_value())
        {
            auto ex = type::Object::DynamicCast(*extends);
            o->setExtends(ex);
        }
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeList(const type::Maybe maybe, const aron::type::VariantPtr& acceptedType)
    {
        auto o = std::make_shared<type::List>(acceptedType);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeDict(const type::Maybe maybe, const aron::type::VariantPtr& acceptedType)
    {
        auto o = std::make_shared<type::Dict>(acceptedType);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writePair(const type::Maybe maybe, const aron::type::VariantPtr& acceptedType1, const aron::type::VariantPtr& acceptedType2)
    {
        auto o = std::make_shared<type::Pair>(acceptedType1, acceptedType2);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeTuple(const type::Maybe maybe, const std::vector<aron::type::VariantPtr>& acceptedTypes)
    {
        auto o = std::make_shared<type::Tuple>(acceptedTypes);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeNDArray(const int ndim, const type::ndarray::ElementType, const type::Maybe maybe)
    {
       auto o = std::make_shared<type::NDArray>();
       return o;
    }

    aron::type::VariantPtr VariantWriter::writeMatrix(const int rows, const int cols, const type::matrix::ElementType type, const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Matrix>();
        o->setMaybe(maybe);
        o->setRows(rows);
        o->setCols(cols);
        o->setElementType(type);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeQuaternion(const type::quaternion::ElementType type, const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Quaternion>();
        o->setMaybe(maybe);
        o->setElementType(type);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeImage(const type::image::PixelType type, const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Image>();
        o->setMaybe(maybe);
        o->setPixelType(type);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writePointCloud(const type::pointcloud::VoxelType type, const type::Maybe maybe)
    {
        auto o = std::make_shared<type::PointCloud>();
        o->setMaybe(maybe);
        o->setVoxelType(type);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writePosition(const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Position>();
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeOrientation(const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Orientation>();
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writePose(const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Pose>();
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeIntEnum(const std::string& name, const std::map<std::string, int>& acceptedValues, const type::Maybe maybe)
    {
        auto o = std::make_shared<type::IntEnum>(name, acceptedValues);
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeInt(const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Int>();
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeLong(const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Long>();
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeFloat(const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Float>();
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeDouble(const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Double>();
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeString(const type::Maybe maybe)
    {
        auto o = std::make_shared<type::String>();
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeBool(const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Bool>();
        o->setMaybe(maybe);
        return o;
    }

    aron::type::VariantPtr VariantWriter::writeTime(const type::Maybe maybe)
    {
        auto o = std::make_shared<type::Time>();
        o->setMaybe(maybe);
        return o;
    }
}
